/*!
 * \file   include/TFEL/Math/General/ResultType.hxx
 * \brief  This file declares the ResultType class.
 * \author Helfer Thomas
 * \date   24 jun 2006
 * \copyright Copyright (C) 2006-2014 CEA/DEN, EDF R&D. All rights 
 * reserved. 
 * This project is publicly released under either the GNU GPL Licence 
 * or the CECILL-A licence. A copy of thoses licences are delivered 
 * with the sources of TFEL. CEA or EDF may also distribute this 
 * project under specific licensing conditions. 
 */

#ifndef _LIB_TFEL_RESULT_TYPE_HXX_
#define _LIB_TFEL_RESULT_TYPE_HXX_

#include"TFEL/Metaprogramming/InvalidType.hxx"
#include"TFEL/Math/General/ComputeObjectTag.hxx"

namespace tfel{

  namespace math{
    
    /*
     * \class ResultType_
     * \brief A helper class for ResultType.
     * This default version returns InvalidType both for
     * Result and Handle.
     * \param typename TagA, tag of the type of the first 
     * argument of the operation.
     * \param typename TagB, tag of the type of second 
     * argument of the operation.
     * \param typename A, type of the first argument of the
     * operation.
     * \param typename B, type of the second argument of the
     * operation.
     * \param typename Op, operation.
     * \return type, type of the result of the operation.
     * \see ResultType.
     * \see ComputeObjectTag.
     */
    template<typename TagA,typename TagB,
	     typename A,   typename B,typename Op>
    struct ResultType_
    {
      /*
       * Result of the binary operation.
       */
      typedef tfel::meta::InvalidType type;

    }; // end of ResultType_

    /*
     * \class ResultType
     * This metafunction gives the result of a binary operation.
     * This class is meant to be specialised on a per objet basis.
     * (in comparison, ComputeBinaryResult is meant to be specialised
     * on a per concept basis)
     * The helper class ResultType_ is only convenient to declare
     * scalar to objet operations.
     * \param typename A, type of the first argument of the
     * operation.
     * \param typename B, type of the second argument of the
     * operation.
     * \param typename Op, operation.
     * \return type, type of the result.
     */
    template<typename A, typename B, typename Op>
    struct ResultType
    {
    private:
      /*
       * Tag of the object A
       */
      typedef typename ComputeObjectTag<A>::type TagA;
      /*
       * Tag of the object A
       */
      typedef typename ComputeObjectTag<B>::type TagB;
    public:
      /*
       * Result
       */
      typedef typename ResultType_<TagA,TagB,A,B,Op>::type type;
    };

  } // end of namespace math

} // end of namespace tfel  

#endif /* _LIB_TFEL_RESULT_TYPE_HXX */

